#include "mesh/DualMeshType.hpp"
#include <utils/pugs_config.hpp>

#ifdef PUGS_HAS_COSTO
#include <language/modules/CouplageModule.hpp>

#include <language/modules/MeshModule.hpp>
#include <language/utils/BuiltinFunctionEmbedder.hpp>
#include <language/utils/SymbolTable.hpp>
#include <language/utils/TypeDescriptor.hpp>
#include <memory>
#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshTraits.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/MeshNodeBoundary.hpp>
#include <mesh/MeshNodeInterface.hpp>
#include <scheme/DirichletBoundaryConditionDescriptor.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/FixedBoundaryConditionDescriptor.hpp>
#include <scheme/IBoundaryConditionDescriptor.hpp>
#include <utils/CommunicatorManager.hpp>
#include <utils/CouplingData.hpp>
#include <utils/Messenger.hpp>
#include <utils/Serializer.hpp>
#ifdef PUGS_HAS_MPI
#include <mpi.h>
#endif   // PUGS_HAS_MPI

CouplageModule::CouplageModule()
{
  this->_addBuiltinFunction("serialize",
                            std::function(
                              [](const std::shared_ptr<const MeshVariant> mesh_v,
                                 const std::vector<std::shared_ptr<const IInterfaceDescriptor>>& i_interface_list,
                                 const int64_t dest,
				 const int64_t tag) -> void {
				return Serializer().interfaces(mesh_v, i_interface_list, dest, tag);
                              }));

  this->_addBuiltinFunction("serialize", std::function([](const std::shared_ptr<const MeshVariant> mesh_v,
							  const int64_t dest,
                                                          const int64_t tag) -> void {
                                return Serializer().mesh(mesh_v, dest, tag);

                            }));

  this->_addBuiltinFunction("serialize", std::function([](const std::shared_ptr<const IBoundaryDescriptor>& boundary,
							  const std::shared_ptr<const MeshVariant> mesh_v,
							  const int64_t location,
                                                          const int64_t dest, const int64_t tag) -> void {
    return Serializer().BC(boundary, mesh_v, location, dest, tag);
  }));


  this->_addBuiltinFunction("cpl_build_graph", std::function([](const std::shared_ptr<const MeshVariant> mesh_v,
								const std::shared_ptr<const IBoundaryDescriptor>& boundary) -> void {
    return Serializer().build_graph(mesh_v, boundary);
  }));

  this->_addBuiltinFunction("cpl_get_dt", std::function([](const double& dt) -> double {
    return Serializer().cpl_get_dt(dt);
  }));

  this->_addBuiltinFunction("serialize_field",
                            std::function([](const std::shared_ptr<const MeshVariant> mesh,
                                             const std::shared_ptr<const IBoundaryDescriptor>& boundary,
                                             const std::shared_ptr<const ItemValueVariant>& field,
					     const int64_t dest,
                                             const int64_t tag) -> void {
			      return Serializer().BC_field(mesh, boundary, field, dest, tag);
                            }));

  // this->_addBuiltinFunction("serialize_field",
  //                           std::function([](const std::shared_ptr<const DiscreteFunctionVariant>& field,
  //                                            const int64_t dest, const int64_t tag) -> void {
  //                             /* const int tag                = 300; */
  //                             // const std::shared_ptr i_mesh = getCommonMesh({field});
  //                             // if (not i_mesh) {
  //                             //   throw NormalError("primal discrete functions are not defined on the same mesh");
  //                             // }

  //                             // switch (i_mesh->dimension()) {
  //                             // case 1: {
  //                             //   return Serializer<1>().field(field, dest, tag);
  //                             // }
  //                             // case 2: {
  //                             //   return Serializer<2>().field(field, dest, tag);
  //                             // }
  //                             // case 3: {
  //                             //   return Serializer<3>().field(field, dest, tag);
  //                             // }
  //                             // default: {
  //                             //   throw UnexpectedError("invalid mesh dimension");
  //                             // }
  //                             // }
  //                           }));

  // this->_addBuiltinFunction("cpl_change_mesh_position",
  //                           std::function([](std::shared_ptr<const MeshVariant> mesh, const int64_t src,
  //                                            const int64_t tag) -> std::shared_ptr<const IMesh> {
  //                             /* tag == 400; */
  //                             /* src = 1 */
  //                             // switch (mesh->dimension()) {
  //                             // case 1: {
  //                             //   return Serializer<1>().change_mesh_position(mesh, src, tag);
  //                             // }
  //                             // case 2: {
  //                             //   return Serializer<2>().change_mesh_position(mesh, src, tag);
  //                             // }
  //                             // case 3: {
  //                             //   return Serializer<3>().change_mesh_position(mesh, src, tag);
  //                             // }
  //                             // default: {
  //                             //   throw UnexpectedError("invalid mesh dimension");
  //                             // }
  //                             // }
  //                           }));

  // this->_addBuiltinFunction("get_flux_and_velocity",
  //                           std::function(

  //                             [](std::shared_ptr<const MeshVariant> mesh,
  //                                const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
  //                                  bc_descriptor_list,
  //                                const std::shared_ptr<const ItemValueVariant>& ur,
  //                                const std::shared_ptr<const SubItemValuePerItemVariant>& Fjr,
  //                                const std::shared_ptr<const ItemValueVariant>& ur_saved,
  //                                const std::shared_ptr<const SubItemValuePerItemVariant>& Fjr_saved)
  //                               -> std::tuple<std::shared_ptr<const ItemValueVariant>,
  //                                             std::shared_ptr<const SubItemValuePerItemVariant>,
  //                                             std::shared_ptr<const ItemValueVariant>,
  //                                             std::shared_ptr<const SubItemValuePerItemVariant>> {
  //                               // switch (i_mesh->dimension()) {
  //                               // case 1: {
  //                               //   auto myCouplingData = CouplingData<1>::getInstanceOrBuildIt();
  //                               //   return myCouplingData.get_flux_and_velocity(i_mesh, bc_descriptor_list, ur, Fjr,
  //                               //                                               ur_saved, Fjr_saved);
  //                               // }
  //                               // case 2: {
  //                               //   auto myCouplingData = CouplingData<2>::getInstanceOrBuildIt();
  //                               //   return myCouplingData.get_flux_and_velocity(i_mesh, bc_descriptor_list, ur, Fjr,
  //                               //                                               ur_saved, Fjr_saved);
  //                               // }
  //                               // case 3: {
  //                               //   auto myCouplingData = CouplingData<3>::getInstanceOrBuildIt();
  //                               //   return myCouplingData.get_flux_and_velocity(i_mesh, bc_descriptor_list, ur, Fjr,
  //                               //                                               ur_saved, Fjr_saved);
  //                               // }
  //                               // default: {
  //                               //   throw UnexpectedError("invalid mesh dimension");
  //                               // }
  //                               // }
  //                               /* } */
  //                               /* } */
  //                             }));

  // this->_addBuiltinFunction("update_mesh", std::function(

  //                                            [](std::shared_ptr<const MeshVariant> mesh,
  //                                               const std::shared_ptr<const IBoundaryDescriptor>& boundary,
  //                                               const int64_t src, const int64_t tag) -> std::shared_ptr<const IMesh> {
  //                                              // switch (i_mesh->dimension()) {
  //                                              // case 1: {
  //                                              //   auto myCouplingData = CouplingData<1>::getInstanceOrBuildIt();
  //                                              //   return myCouplingData.update_mesh(i_mesh, boundary, src, tag);
  //                                              // }
  //                                              // case 2: {
  //                                              //   auto myCouplingData = CouplingData<2>::getInstanceOrBuildIt();
  //                                              //   return myCouplingData.update_mesh(i_mesh, boundary, src, tag);
  //                                              // }
  //                                              // case 3: {
  //                                              //   auto myCouplingData = CouplingData<3>::getInstanceOrBuildIt();
  //                                              //   return myCouplingData.update_mesh(i_mesh, boundary, src, tag);
  //                                              // }
  //                                              // default: {
  //                                              //   throw UnexpectedError("invalid mesh dimension");
  //                                              // }
  //                                              // }
  //                                            }));

  this->_addBuiltinFunction("hello", std::function(

                                       []() -> void { std::cout << "in test" << std::endl; }

                                       ));

}

void
CouplageModule::registerOperators() const
{}

/* void */
/* CouplageModule::_test() */
/* { */
/*   std::cout << "in test" << std::endl; */
/* } */

#endif   // PUGS_HAS_COSTO
