#ifndef GNUPLOT_WRITER_1D_HPP
#define GNUPLOT_WRITER_1D_HPP

#include <output/WriterBase.hpp>

#include <algebra/TinyMatrix.hpp>
#include <algebra/TinyVector.hpp>
#include <mesh/MeshTraits.hpp>
#include <output/OutputNamedItemValueSet.hpp>

class MeshVariant;

#include <optional>
#include <string>

class GnuplotWriter1D final : public WriterBase
{
 private:
  std::string _getDateAndVersionComment() const;

  std::string _getFilename() const;

  template <typename ItemDataType>
  bool _is_cell_data(const ItemDataType&) const;

  template <typename ItemDataType>
  bool _is_face_data(const ItemDataType&) const;

  template <typename ItemDataType>
  bool _is_edge_data(const ItemDataType&) const;

  template <typename ItemDataType>
  bool _is_node_data(const ItemDataType&) const;

  template <typename DataType, ItemType item_type>
  size_t _itemDataNbRow(const ItemValue<DataType, item_type>&) const;

  template <typename DataType, ItemType item_type>
  size_t _itemDataNbRow(const ItemArray<DataType, item_type>&) const;

  template <MeshConcept MeshType, ItemType item_type>
  void _writeItemDatas(const MeshType& mesh,
                       const OutputNamedItemDataSet& output_named_item_data_set,
                       std::ostream& fout) const;

  void _writePreamble(const OutputNamedItemDataSet& output_named_item_value_set, std::ostream& fout) const;

  template <MeshConcept MeshType>
  void _write(const MeshType& mesh,
              const OutputNamedItemDataSet& output_named_item_value_set,
              std::optional<double> time) const;

  void _writeAtTime(const MeshVariant& mesh_v,
                    const std::vector<std::shared_ptr<const INamedDiscreteData>>& named_discrete_data_list,
                    double time) const final;

  void _write(const MeshVariant& mesh_v,
              const std::vector<std::shared_ptr<const INamedDiscreteData>>& named_discrete_data_list) const final;

  void _writeMesh(const MeshVariant& mesh_v) const final;

 public:
  GnuplotWriter1D(const std::string& base_filename) : WriterBase(base_filename) {}

  GnuplotWriter1D(const std::string& base_filename, const double time_period) : WriterBase(base_filename, time_period)
  {}

  ~GnuplotWriter1D() = default;
};

#endif   // GNUPLOT_WRITER_1D_HPP
