#pragma once
#include <utils/pugs_config.hpp>
#include <cstdint>

#ifdef PUGS_HAS_COSTO

#include <language/utils/InterpolateItemValue.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/ItemValueVariant.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>
#include <mesh/MeshTraits.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/MeshFaceBoundary.hpp>
#include <mesh/MeshFlatFaceBoundary.hpp>
#include <mesh/MeshNodeBoundary.hpp>
#include <mesh/MeshNodeInterface.hpp>
#include <scheme/DirichletBoundaryConditionDescriptor.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/NeumannBoundaryConditionDescriptor.hpp>
#include <scheme/SymmetryBoundaryConditionDescriptor.hpp>
#include <string>
#include <type_traits>
#include <utils/FPEManager.hpp>
#include <utils/Messenger.hpp>
#include <utils/PugsAssert.hpp>
#include <utils/PugsMacros.hpp>
#include <utils/pugs_config.hpp>
#include <vector>

#ifdef PUGS_HAS_MPI
#include <mpi.h>
#endif   // PUGS_HAS_MPI
class MeshVariant;

class Serializer
{
private:
  // std::shared_ptr<const MeshVariant> m_mesh_v;

  // std::shared_ptr<const MeshVariant> m_mesh;
  // static constexpr size_t Dimension = MeshType::Dimension;
  // using Rd           = TinyVector<Dimension>;
  // using MeshDataType = MeshData<MeshType>;

public:
  enum locator : int
    {
      at_node,
      at_face,
      at_cell_center,
    };
  Serializer() = default;
  // Serializer(const std::shared_ptr<const MeshVariant>& mesh_v) : m_mesh_v{mesh_v} {}

  // ~Serializer() = dafaults;
  void mesh(const std::shared_ptr<const MeshVariant> mesh_v,
	    const int64_t dest,
	    const int64_t tag);



  void build_graph(const std::shared_ptr<const MeshVariant> mesh_v,
		   const std::shared_ptr<const IBoundaryDescriptor>& boundary)
  {
    this->build_graph(mesh_v,*boundary);
  }

  void build_graph(const std::shared_ptr<const MeshVariant> mesh_v,
		   const IBoundaryDescriptor& boundary);

  double cpl_get_dt(const double& dt);


  void
  BC(const std::shared_ptr<const IBoundaryDescriptor>& boundary,
     const std::shared_ptr<const MeshVariant> mesh_v,
     const int location,
     const int64_t dest,
     const int64_t tag)
  {
    this->BC(*boundary, mesh_v, location, dest, tag);
  }

  void BC(const IBoundaryDescriptor& boundary,
	  const std::shared_ptr<const MeshVariant> mesh_v,
          const int location,
          const int64_t dest,
          const int64_t tag);


  void BC_field(const std::shared_ptr<const MeshVariant> mesh_v,
                const std::shared_ptr<const IBoundaryDescriptor>& boundary,
                const std::shared_ptr<const ItemValueVariant>& field,
                const int64_t dest,
                const int64_t tag);

  void interfaces(const std::shared_ptr<const MeshVariant> mesh_v,
                  const std::vector<std::shared_ptr<const IInterfaceDescriptor>>& i_interface_list,
                  const int64_t dest,
                  const int64_t tag);

  // void field(const std::shared_ptr<const DiscreteFunctionVariant>& field, const int64_t dest, const int64_t tag);

  // std::shared_ptr<const MeshVariant> change_mesh_position(std::shared_ptr<const MeshVariant> mesh_v,
  // 							  const int64_t src,
  // 							  const int64_t tag);
};





// // Field (faces)
// template <MeshConcept MeshType>void
// Serializer<MeshType>::field(const std::shared_ptr<const DiscreteFunctionVariant>& field,
//                              const int64_t dest,
//                              const int64_t tag)
// {
//   const std::shared_ptr i_mesh = getCommonMesh({field});
//   if (not i_mesh) {
//     throw NormalError("primal discrete functions are not defined on the same mesh");
//   }
//   /* assert(MeshType(i_mesh) == field->mesh()); */
//   const std::shared_ptr mesh = std::dynamic_pointer_cast<const MeshType>(i_mesh);
//   /* const IBoundaryDescriptor& bd = *boundary; */
//   /* const Mesh<Connectivity<2>> mt = <const MeshType>(i_mesh); */

//   /* const auto& n_list = MeshNodeBoundary<2>(mt, bd); */

//   const NodeValue<const Rd>& xr = mesh->xr();
//   Array<TinyVector<Dimension>> positions(mesh->numberOfNodes());
//   parallel_for(
//     mesh->numberOfNodes(), PUGS_LAMBDA(NodeId r) {
//       for (unsigned short i = 0; i < Dimension; ++i) {
//         positions[r][i] = xr[r][i];
//       }
//       /* for (unsigned short i = Dimension; i < 3; ++i) { */
//       /*   positions[r][i] = 0; */
//       /* } */
//     });

//   std::vector<double> pts;
//   pts.resize(mesh->numberOfNodes() * Dimension);
//   for (unsigned short r = 0; r < mesh->numberOfNodes(); ++r) {
//     for (unsigned short j = 0; j < Dimension; ++j) {
//       pts[Dimension * r + j] = positions[r][j];
//     }
//     /* std::cout << "\033[01;31m" << pts[2 * r] << "; " << pts[2 * r + 1] << "\033[00;00m" << std::endl; */
//   }

//   auto costo = parallel::Messenger::getInstance().myCoupling;
//   /* const int64_t dest = costo->myGlobalSize() - 1; */
//   std::vector<int> shape;
//   shape.resize(3);
//   shape[0] = mesh->numberOfNodes();
//   shape[1] = Dimension;
//   shape[2] = 0;
//   costo->sendData(shape, pts, dest, tag);

//   std::vector<int> recv;
//   const int src = costo->myGlobalSize() - 1;

//   costo->recvData(recv, src, tag + 1);
// }

// template <MeshConcept MeshType>
// std::shared_ptr<const MeshVariant>
// Serializer<MeshType>::change_mesh_position(std::shared_ptr<const MeshVariant> mesh_v, const int64_t src, const int64_t tag)
// {
//   std::vector<double> recv_xr;
//   parallel::Messenger::getInstance().myCoupling->recvData(recv_xr, src, tag);

//   const std::shared_ptr given_mesh = std::dynamic_pointer_cast<const MeshType>(mesh_v);

//   NodeValue<Rd> new_xr{given_mesh->connectivity()};
//   parallel_for(
//     given_mesh->numberOfNodes(), PUGS_LAMBDA(const NodeId node_id) {
//       for (unsigned short j = 0; j < Dimension; ++j) {
//         /* new_xr[node_id][j] = xr[node_id][j]; */
//         new_xr[node_id][j] = recv_xr[Dimension * node_id + j];
//       }
//     });

//   return std::make_shared<MeshVariant>(given_mesh->shared_connectivity(), new_xr);
// }

// using Serializer3D = Serializer<Mesh<1>>;
// using Serializer2D = Serializer<Mesh<2>>;
// using Serializer1D = Serializer<Mesh<3>>;

#endif   // PUGS_HAS_COSTO
// #endif   // SERIALIZER_HPP
